
from bellameta import constants
from bellameta.utils import get_config, custom_title

class BellametaType:
    ''' 
    A class representing the base metadata type serving as a blue print for child classes.

    Attributes
    ----------
    The attributes are dynamically set by the inheriting class.
    
    Methods
    -------
    from_str: 
        create class by providing a string (that is registered in the inheriting classe's value dict)
    from_int:
        create class by providing an integer (that is registered in the inheriting classe's value dict)
    list:
        list all values registered in the inheriting classe's value dict
    to_string:
        convert class into string (registered in inherting classe's value dict)
    
    '''

    def __new__(cls, s: str):
        return cls.from_str(s)
    
    def __init_subclass__(cls, values):
        super().__init_subclass__()
        if not hasattr(cls, 'values') or values:
            cls.values = values
            cls._key_map = {}
            for value in list(values.keys()):
                attr_name = custom_title(value.replace(" ", "_"))
                attr_value = type(f"{cls.__name__}{attr_name}", (), {
                    'to_string': classmethod(lambda cls, original_key=value: original_key),
                    # TODO: we do not use this anymore
                    'to_table_name': classmethod(lambda cls, original_key=value: original_key.lower())
                })
                # this is needed for proper inheritance, i.e., that e.g. Task.Subtyping also gets the labels class attribute
                for attr, val in cls.__dict__.items():
                    if not attr.startswith('__') and attr not in ['values', '_key_map'] and not callable(val):
                        setattr(attr_value, attr, val)
                setattr(cls, attr_name, attr_value)
                cls._key_map[attr_name] = value
    
    @classmethod
    def from_str(cls, s: str):
        clean = s.replace(" ", "_").title()
        if clean in vars(cls).keys():
            return getattr(cls, clean)
        else:
            raise ValueError(f"{s} is not a valid attribute.")     

        
    @classmethod
    def from_int(cls, i: int):
        value = list(cls.values.keys())[list(cls.values.values()).index(i)]
        return getattr(cls, value)
    
    @classmethod
    def list(cls):
        return list(cls.values.keys())

# Each new type can be specified via a new entry in the config_data yaml
class Cohort(BellametaType, values=constants.COHORTS):
    pass

class Task(BellametaType, values=constants.TASKS):
    pass




from os.path import join as _j
from typing import List, Tuple
import sqlite3 
import pandas as pd

from bellameta.utils import sqlite_connection
from bellameta import constants



def create_statement(table_name: str, type = 'TEXT'):
    statement = f'''CREATE TABLE {table_name} (
                hash TEXT,
                value {type},
                UNIQUE(hash, value),
                FOREIGN KEY(hash) REFERENCES storage(hash)
                )'''
    return statement

class Db():
    '''
    Class to manage metadata within sqlite table

    Attributes
    ----------
    sqlite_path: str
        Absolute path to the sqlite database holding storage table of WSI scans
    tables: List[str]
        Names of metadata tables to be initialized/modified

    Methods
    -------
    write:
        Method to write metadata as key (hash) value pair into the specified metadata table
    write_many:
        Bulk insert of metadata into the specified metadata table
    '''

    def __init__(self, tables = ['state', 'cohort', 'patient', 'section', 'tag', 'stain', 'task', 'subtype', 'year', 'gleason_grade']):
        self.sqlite_path = constants.DB_PATH
        self.tables = tables
        self._initialize_db()

    @sqlite_connection
    def _initialize_db(self, cursor):
        existing_tables = []
        for table_name in self.tables:
            if not self._table_exists(table_name):
                print(f'Creating table {table_name}')
                cursor.execute(create_statement(table_name))
            else:
                existing_tables.append(table_name)
        print(f'Already existing metdata tables: {existing_tables}')

    @sqlite_connection
    def _table_exists(self, cursor, table_name: str):
        cursor.execute(
            "SELECT name FROM sqlite_master WHERE name=?", (table_name, ))
        result = cursor.fetchall()
        if result:
            return True
        else:
            False


    @sqlite_connection
    def drop_all_metadata_tables(self, cursor):
        '''
        This is just a utility function that should normally not be used
        '''

        for table in self.tables:
            cursor.execute(f'''DELETE FROM {table}''')
                
    @sqlite_connection
    def write(self, cursor, hash: str, value: str, table_name: str):
        '''
        Method to write metadata as key (hash) value pair into the specified metadata table

        Parameters
        ----------
        hash : str
            Hash of WSI scan, serving as unique key in the database
        value : str
            metadata value
        table_name : str
            name of the metadata table to be inserted to
        '''

        if value:
            cursor.execute(
                f"INSERT OR IGNORE INTO {table_name} (hash, value) VALUES (?, ?)",
                (hash, value)
            )
    @sqlite_connection
    def write_many(self, cursor, data: List[Tuple], table_name):
        '''
        Bulk insert of metadata into the specified metadata table
        '''

        # validation
        filtered_data = [(hash, value) for (hash, value) in data if value is not None]
        cursor.executemany(
            f"INSERT OR IGNORE INTO {table_name} (hash, value) VALUES (?, ?)",
            filtered_data
        )
    
    def read_pd(self, table_name: str, rows: int | None = 5):
        '''
        Reads in the first rows of the table as pandas dataframe 
        '''

        conn = sqlite3.connect(self.sqlite_path)
        statement = f"SELECT * FROM {table_name}"
        if rows is not None:
            statement += f" LIMIT {rows}"
        df = pd.read_sql_query(statement, conn)
        conn.close()
        return df
    
    def __str__(self):
        out = ""
        for table_name in self.tables:
            df = self.read_pd(table_name, rows=5)
            out += f"Table {table_name}:\n {df.to_string()}\n"
        return(out)
    

                
            




from typing import List
from abc import ABC, abstractmethod
from pamly import Stain

from bellameta.utils import sqlite_connection
from bellameta.database import Db
from bellameta.types import Cohort, Task

class Metadata(ABC):
    '''
    Class that automates handling of metadata implemented in its child classes

    The beauty of this class is its dynamic definition of getter methods,
    based on metadata table names and the logic implemented in each child class.
    Each child class has to implement getters for the tables cohort, stain and task.

    Attributes
    ----------
    db: Db
        Database holding scans and metadata
    ingress_directory: str
        Absolute path of the original ingress directory that held scans BEFORE moving into storage.
        This is used to identify all scans of a single cohort.
    
    Methods
    -------
    get_hashes:
        Fetches all hashes of scans contained in original ingress directory (we think of this as a single cohort)
    get_scanname_from_hash:
        Returns the original scan name (not the hash) as specified by the clinic as stored in the storage table.
        Usually the scanname holds valuable metadata such as patient id, subtype, ... so this is the main source
        of metadata for further processing.
    write_many:
        Extends Db.write_many by using the dedicated getter methods of child class in order to preprocess data to be inserted
        into metadata tables.
    '''

    def __init__(self, db: Db, ingress_directory: str):

        self.db = db
        self.sqlite_path = db.sqlite_path
        self.ingress_directory = ingress_directory
        
        # This sets up getter methods for each table, returning None if not overwritten in child class
        def __getattr__(self, name):
            # Check if this is a get_* method
            if name.startswith('get_'):
                table = name[4:]  # Remove 'get_' prefix
                if table in self.db.tables:
                    return lambda hash, table=table: self._get_record(hash, table)
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")
        
        if '__getattr__' not in self.__class__.__dict__:
            setattr(self.__class__, '__getattr__', __getattr__)

    def _get_record(self, hash, table):
        return None
    
    # this forces the user to implement the three main getter methods for cohort, stain and task
    @abstractmethod
    def get_cohort(self, hash) -> List[Cohort] | Cohort:
        '''
        If a list is returned a hash will have multiple entries in the cohort table.
        '''

        pass

    @abstractmethod
    def get_stain(self,hash) -> Stain:
        pass

    @abstractmethod
    def get_task(self, hash) -> List[Task] | Task:
        '''
        If a list is returned a hash will have multiple entries in the task table.
        '''

        pass

    @sqlite_connection
    def get_hashes(self, cursor):
        '''
        Fetches all hashes of scans contained in original ingress directory (we think of this as a single cohort)
        '''

        cursor.execute(f'''
                        SELECT hash FROM ingress
                        WHERE filepath GLOB '{self.ingress_directory}/*'
                       ''')
        data = cursor.fetchall()
        hashes = [entry[0] for entry in data]
        return hashes
    
    @sqlite_connection
    def get_scanname_from_hash(self, cursor, hash: str):
        '''
        Returns the original scan name (not the hash) as specified by the clinic as stored in the storage table.

        Usually the scanname holds valuable metadata such as patient id, subtype, ... so this is the main source
        of metadata for further processing.
        '''

        cursor.execute('''
                SELECT scanname FROM storage
                WHERE hash=?
                ''', (hash,))
        data = cursor.fetchone()
        return data[0]     
    
    def write_many(self):
        '''
        Extends Db.write_many by using the dedicated getter methods of child class in order to preprocess data to be inserted
        into metadata tables.

        This inserts all metadata into all metadata tables for which the child class
        implements the respective getter methods.

        If getter method returns list, there will be multiple entries for a hash in the respective metadata table
        '''

        hashes = self.get_hashes()
        for table in self.db.tables:
            method_name = f'get_{table}'
            get_method = getattr(self, method_name)
            data = []
            for hash in hashes:
                values = get_method(hash)
                if isinstance(values, List):
                    for value in values:
                        data.append((hash, value))
                else:
                    data.append((hash, values))
            self.db.write_many(data, table)
        


import os
from os.path import join as _j
import sqlite3
from typing import List
import functools
import pandas as pd
from dotenv import load_dotenv
from ruamel.yaml import YAML

# DATABSES
def sqlite_connection(func):
    '''
    Decorator to facilitate sqlite connection

    Wraps a function that creates a cursor, commits transactions and finally closes the connection
    '''
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        with sqlite3.connect(self.sqlite_path) as conn:
            cursor = conn.cursor()
            try:
                result = func(self, cursor, *args, **kwargs)
                conn.commit()
                return result
            except Exception as e:
                conn.rollback()
                raise e
    return wrapper

def get_env(env_path: str|None = None):
    '''
    Searches for a .env file and loads it.

    Parameters
    ----------
    env_path : str|None
        Path to the .env file. If stored at the root of the repository
        load_dotenv will find it automatically.

    Raises
    ------
        RuntimeError if no .env file is found

    Return
    ------
        True if .env file was found
    '''

    success = load_dotenv(env_path)
    if not success:
        raise RuntimeError("No valid .env file found. Make sure to create a .env file at the root of the repository defining CONFIG_PATH.")
    return success

def get_config(config_path: str):
    '''
    Gets the config stored in a .yaml file under the path specified in the .env file
    Parameters
    ----------
    env_path : str|None
        Path to the .env file. If stored at the root of the repository
        load_dotenv will find it automatically.
    
    Return
    ------
        Contents of the yaml file
    '''
    if not os.path.exists(config_path):
        raise RuntimeError("Specified path does not exists. Please define CONFIG_PATH in .env to be a path to a yaml config.")
    yaml = YAML(typ="rt")
    yaml.preserve_quotes = True
    with open(config_path, "r") as f:
        config_data = yaml.load(f)
        return config_data

def custom_title(text):
    '''
    Transform a possibly mispelled text into a consistent title
    '''

    result = []
    make_upper = True
    
    for char in text:
        if char.isalpha():
            if make_upper:
                result.append(char.upper())
            else:
                # Preserve existing case
                result.append(char)
            make_upper = False
        else:
            result.append(char)
            make_upper = True
            
    return ''.join(result)